/* Copyright Robin Schirrmeister and Simon Skilevic
 * Lecture Efficient Route Planning summer semester 2011
 * Uni Freiburg
 */

#include "./Dijkstra.h"
#include <utility>
#include <algorithm>
#include <vector>
#include <sstream>
#include <iostream>
#include <queue>
#include <string>
#include <functional>
#include <algorithm>
#include "./Timer.h"

using std::pair;
using std::cout;
using std::endl;
// _____________________________________________________________________________
Dijkstra::Dijkstra(Graph* roadNetwork,
                   DistanceHeuristic* heuristic) :_graph(roadNetwork)
{
  this->heuristic = heuristic;
  resetAll();
}

Dijkstra::~Dijkstra()
{
}

void Dijkstra::resetAll()
{
  _openNodeQueue = std::priority_queue <pair<double, int>,
                    vector<pair<double, int> >,
                    std::greater<pair<double, int > > >();
  size_t graphSize = _graph->getNumNodes();
  weights.clear();
  heuristicValues.clear();
  timesFromStart.clear();
  expandedFrom.clear();
  expanded.clear();
  weights.resize(graphSize, 0.0);
  heuristicValues.resize(graphSize, -1.0);
  timesFromStart.resize(graphSize, 0.0);
  pair<int, int> idEdgeIdPair(-1, -1);
  expandedFrom.resize(graphSize, idEdgeIdPair);
  expanded.resize(graphSize, false);
  exploredNodes = 0;
}

void Dijkstra::setHeuristic(DistanceHeuristic* heuristic)
{
  this->heuristic = heuristic;
}

template<bool checkNodes, bool checkEdges, bool rememberEdges,
         bool forward, bool toAllNodes>
double Dijkstra::computeShortestPath(int startId, int endId)
{
  // TODO(Robin): make reset optional... move it out of this function...
  atStartOfDijkstra();
  NodeExpanded* startNode = _graph->getNode(startId);
  NodeExpanded* endNode;
  if (!toAllNodes)
    endNode = _graph->getNode(endId);
  timesFromStart[startId] = 0;
  if (!toAllNodes)
  {
    heuristicValues[startId] = heuristic->approximateDistance(startNode,
                                                              endNode);
    weights[startId] = heuristicValues[startId];
  }
  _openNodeQueue.push(pair<double, int>(weights[startId], startId));
  int currentId;
  vector<Edge>* edges;

  int newId;
  double newWeight;
  bool dijkstraFinished = false;
  _stopDijkstra = false;
  // get straight line distance to endnode
  while (!dijkstraFinished && !_stopDijkstra)
  {
    // get minimum of node queue
    currentId = _openNodeQueue.top().second;
    _openNodeQueue.pop();

    // check if already explored
    if (expanded[currentId] == true)
    {
      // TODO(Robin): only check if queue is empty?
      // because other case can't happen ?
      if (toAllNodes)
          dijkstraFinished = _openNodeQueue.empty();
      else
          dijkstraFinished = currentId == endId || _openNodeQueue.empty();
       continue;
    }
    // check if this node is allowed by seleceted algorithm
    if (checkNodes)
    {
      if (!nodeAllowed(currentId))
      {
        // reset node, then continue...
        resetNode(currentId);
        if (toAllNodes)
          dijkstraFinished = _openNodeQueue.empty();
        else
          dijkstraFinished = currentId == endId || _openNodeQueue.empty();
        continue;
      }
    }
    // get outgoing edges from this node
    if (forward)
      edges = _graph->getOutEdges(currentId);
    else
      edges = _graph->getInEdges(currentId);

    // compute all new distances in loop
    for (size_t i = 0; i < edges->size(); i++)
    {
      // get all new distances
      Edge& newEdge = (*edges)[i];
      if (checkEdges)
      {
        if (!edgeAllowed(forward, currentId, i)) continue;
      }
      newId = newEdge.otherNodeId;

      // if node has already been explored = expanded, ignore it...
      if (expanded[newId] == true)
        continue;

      // compute heuristic value of new node if necessary
      if (heuristicValues[newId] == -1)
      {
        NodeExpanded* newNode =  _graph->getNode(newId);
        if (toAllNodes)
          heuristicValues[newId] = 0;
        else
          heuristicValues[newId] =
                  heuristic->approximateDistance(newNode, endNode);
        timesFromStart[newId] = timesFromStart[currentId] + newEdge.time;
        weights[newId] = timesFromStart[newId] + heuristicValues[newId];
        _openNodeQueue.push(pair<double, int>(weights[newId], newId));
        if (rememberEdges)
            expandedFrom[newId] = pair<int, int> (currentId, i);
      }
      else
      {
        newWeight = timesFromStart[currentId] + newEdge.time +
                    heuristicValues[newId];
        if (newWeight < weights[newId])
        {
          timesFromStart[newId] = timesFromStart[currentId] + newEdge.time;
          weights[newId] = timesFromStart[newId] + heuristicValues[newId];
          _openNodeQueue.push(pair<double, int>(weights[newId], newId));
          if (rememberEdges)
            expandedFrom[newId] = pair<int, int> (currentId, i);
        }
      }
    }
    expanded[currentId] = true;
    resetNode(currentId);
    exploredNodes++;
    // TODO(Robin): make optinal resetnode function here,
    // also reset nodes left in queue at end..
    if (toAllNodes)
      dijkstraFinished = _openNodeQueue.empty();
    else
      dijkstraFinished = currentId == endId || _openNodeQueue.empty();
  }
  int lastSettledId = currentId;
  // reset all ndos left in queue, empty queue this way...
  while (!_openNodeQueue.empty())
  {
    currentId = _openNodeQueue.top().second;
    resetNode(currentId);
    _openNodeQueue.pop();
  }
  assert(_openNodeQueue.empty());
  if (endId != lastSettledId && !toAllNodes)
  {
    printf("shortest Path not found from %d to %d\n", startId, endId);
    return -1;
  }
  return timesFromStart[endId];
}

double Dijkstra::computeShortestPath(int startId, int endId,
        bool checkNodes, bool checkEdges, bool rememberEdges, bool forward)
{
  if      (checkNodes && checkEdges  && rememberEdges  && forward)
    return computeShortestPath<true, true,  true,
                               true,  false> (startId, endId);
  else if (checkNodes && checkEdges  && rememberEdges  && !forward)
    return computeShortestPath<true, true,  true,
                               false, false> (startId, endId);
  else if (checkNodes && checkEdges  && !rememberEdges && forward)
    return computeShortestPath<true, true,  false,
                               true,  false> (startId, endId);
  else if (checkNodes && checkEdges  && !rememberEdges && !forward)
    return computeShortestPath<true, true,  false,
                              false, false> (startId, endId);
  else if (checkNodes && !checkEdges && rememberEdges  && forward)
    return computeShortestPath<true, false, true,
                               true,  false> (startId, endId);
  else if (checkNodes && !checkEdges && rememberEdges  && !forward)
    return computeShortestPath<true, false, true,
                               false, false> (startId, endId);
  else if (checkNodes && !checkEdges && !rememberEdges && forward)
    return computeShortestPath<true, false, false,
                               true,  false> (startId, endId);
  else if (checkNodes && !checkEdges && !rememberEdges && !forward)
    return computeShortestPath<true, false, false,
                               false, false> (startId, endId);
  // now do not check nodes..
  else if (!checkNodes && checkEdges  && rememberEdges  && forward)
    return computeShortestPath<false, true,  true,
                               true,  false> (startId, endId);
  else if (!checkNodes && checkEdges  && rememberEdges  && !forward)
    return computeShortestPath<false, true,  true,
                               false, false> (startId, endId);
  else if (!checkNodes && checkEdges  && !rememberEdges && forward)
    return computeShortestPath<false, true,  false,
                               true,  false> (startId, endId);
  else if (!checkNodes && checkEdges  && !rememberEdges && !forward)
    return computeShortestPath<false, true,  false,
                              false, false> (startId, endId);
  else if (!checkNodes && !checkEdges && rememberEdges  && forward)
    return computeShortestPath<false, false, true,
                               true,  false> (startId, endId);
  else if (!checkNodes && !checkEdges && rememberEdges  && !forward)
    return computeShortestPath<false, false, true,
                               false, false> (startId, endId);
  else if (!checkNodes && !checkEdges && !rememberEdges && forward)
    return computeShortestPath<false, false, false,
                               true,  false> (startId, endId);
  else if (!checkNodes && !checkEdges && !rememberEdges && !forward)
    return computeShortestPath<false, false, false,
                               false, false> (startId, endId);
  else
  {
    printf("Unbekannte aufrufparameter bei compute shortest path (Dijkstra.cpp)"
            "\n");
    return -1;
  }
}
double Dijkstra::computeShortestPathToAll(int startId,
                                          bool checkNodes, bool checkEdges,
                                          bool rememberEdges, bool forward)
{
  int endId = -1;

  if      (checkNodes && checkEdges  && rememberEdges  && forward)
    return computeShortestPath<true, true,  true,
                               true,  true> (startId, endId);
  else if (checkNodes && checkEdges  && rememberEdges  && !forward)
    return computeShortestPath<true, true,  true,
                               false, true> (startId, endId);
  else if (checkNodes && checkEdges  && !rememberEdges && forward)
    return computeShortestPath<true, true,  false,
                               true,  true> (startId, endId);
  else if (checkNodes && checkEdges  && !rememberEdges && !forward)
    return computeShortestPath<true, true,  false,
                              false, true> (startId, endId);
  else if (checkNodes && !checkEdges && rememberEdges  && forward)
    return computeShortestPath<true, false, true,
                               true,  true> (startId, endId);
  else if (checkNodes && !checkEdges && rememberEdges  && !forward)
    return computeShortestPath<true, false, true,
                               false, true> (startId, endId);
  else if (checkNodes && !checkEdges && !rememberEdges && forward)
    return computeShortestPath<true, false, false,
                               true,  true> (startId, endId);
  else if (checkNodes && !checkEdges && !rememberEdges && !forward)
    return computeShortestPath<true, false, false,
                               false, true> (startId, endId);
  // now do not check nodes..
  else if (!checkNodes && checkEdges  && rememberEdges  && forward)
    return computeShortestPath<false, true,  true,
                               true,  true> (startId, endId);
  else if (!checkNodes && checkEdges  && rememberEdges  && !forward)
    return computeShortestPath<false, true,  true,
                               false, true> (startId, endId);
  else if (!checkNodes && checkEdges  && !rememberEdges && forward)
    return computeShortestPath<false, true,  false,
                               true,  true> (startId, endId);
  else if (!checkNodes && checkEdges  && !rememberEdges && !forward)
    return computeShortestPath<false, true,  false,
                              false, true> (startId, endId);
  else if (!checkNodes && !checkEdges && rememberEdges  && forward)
    return computeShortestPath<false, false, true,
                               true,  true> (startId, endId);
  else if (!checkNodes && !checkEdges && rememberEdges  && !forward)
    return computeShortestPath<false, false, true,
                               false, true> (startId, endId);
  else if (!checkNodes && !checkEdges && !rememberEdges && forward)
    return computeShortestPath<false, false, false,
                               true,  true> (startId, endId);
  else if (!checkNodes && !checkEdges && !rememberEdges && !forward)
    return computeShortestPath<false, false, false,
                               false, true> (startId, endId);
  else
  {
    printf("Unbekannte aufrufparameter bei compute shortest path to all "
           "(Dijkstra.cpp) \n");
    return -1;
  }
}
SimpleDijkstra::SimpleDijkstra(Graph* roadNetwork,
                   DistanceHeuristic* heuristic)
                  : Dijkstra(roadNetwork, heuristic)
{
}
void SimpleDijkstra::doPrecomputation() { }

void SimpleDijkstra::atStartOfDijkstra()
{
  // check if graph size changed, else just reset expandedvector...
  if (weights.size() != _graph->getNumNodes())
  {
    resetAll();
  }
  else
  {
    exploredNodes = 0;
    expanded.clear();
    expanded.resize(_graph->getNumNodes(), false);
  }
}

void SimpleDijkstra::resetNode(int nodeId)
{
  heuristicValues[nodeId] = -1;
  weights[nodeId] = 0.0;
}
bool SimpleDijkstra::edgeAllowed(bool outEdge, int nodeId, int edgeId)
{ return true; }

bool SimpleDijkstra::nodeAllowed(int nodeId)
{ return true; }

double SimpleDijkstra::computeShortestPath(int startId, int endId,
                                          bool reconstructPath, bool forward)
{
  bool checkNodes = false;
  bool checkEdges = false;
  double result = Dijkstra::computeShortestPath(startId, endId, checkNodes,
                                          checkEdges, reconstructPath, forward);
  if (reconstructPath)
  {
    if (forward)
      reconstructShortestPath(startId, endId);
    else
      reconstructShortestPath(endId, startId);
  }
  return result;
}

double SimpleDijkstra::computeShortestPath(int startId, int endId,
                                                  bool reconstructPath)
{
  return computeShortestPath(startId, endId, reconstructPath, true);
}

void SimpleDijkstra::reconstructShortestPath(int startId, int endId)
{
  shortestPath.clear();
  // check if dijkstra finished...
  if (expanded[endId] == false)
    return;
  // now insert shortest Path into vector
  int nextNodeId = endId;

  while (startId != nextNodeId)
  {
    shortestPath.push_back(*(_graph->getNode(nextNodeId)));
    nextNodeId = expandedFrom[nextNodeId].first;
  }
  // push back start id..
  shortestPath.push_back(*(_graph->getNode(nextNodeId)));
  // have to reverse the path to get proper path
  reverse(shortestPath.begin(), shortestPath.end());
}
