// Copyright Robin Schirrmeister and Simon Skilevic
// University Freiburg 2011

#include "./ExpandedDijkstra.h"
#include <algorithm>
#include <vector>

ExpandedDijkstra::ExpandedDijkstra(TransitNetworkExpanded* transitNetwork)
{
  this->_transitGraph = transitNetwork;
  _expandedFrom.resize(transitNetwork->getNumNodes());
  _timesFromStart.resize(_transitGraph->getNumNodes(), -1);
}

ExpandedDijkstra::~ExpandedDijkstra()
{

}

double ExpandedDijkstra::computeShortestPath(string startStop, Time time,
                                      string endStop, bool buildPath)

{
  return computeShortestPath(_transitGraph->getStopId(startStop), time,
                      _transitGraph->getStopId(endStop), buildPath);
}

double ExpandedDijkstra::computeShortestPath(int startStopId, Time startTime,
                                           int endStopId, bool reconstructPath)
{
  transitShortestPath.clear();
  assert(_timesFromStart.size() == _transitGraph->getNumNodes());
  assert(_expandedFrom.size() == _transitGraph->getNumNodes());
  // reset vectors for algorithm
  _expanded.clear();
  //_timesFromStart.clear();
  _expanded.resize(_transitGraph->getNumNodes(), false);
 // _timesFromStart.resize(_transitGraph->getNumNodes(), -1);
  // get the start node ids
  vector<int> startNodeIds;
  _transitGraph->getStartNodeIds(startStopId, startTime, &startNodeIds);

  // Cannot do Dijkstra without start ids..
  if (startNodeIds.size() == 0)
  {
    return -1;
  }
  // empty priority queue, then enter the starting nodes into the priority queue
  _openNodeQueue = std::priority_queue <pair<int, int>,
            vector<pair<int, int> >,
            std::greater<pair<int, int > > >();
  int waitTimeToNode;
  Time nodeTime;
  for (size_t i = 0; i < startNodeIds.size(); i++)
  {
    nodeTime = _transitGraph->getNode(startNodeIds[i])->time;
    waitTimeToNode = nodeTime.minusInSecondsDependent(startTime);

    _timesFromStart[startNodeIds[i]] = waitTimeToNode;
    _openNodeQueue.push(std::pair<int, int> (waitTimeToNode,
                                                startNodeIds[i]));
  }
  int currentId;
  vector<Edge>* edges;

  int newId;
  int newTimeFromStart;
  bool dijkstraFinished = false;
  size_t exploredNodes = 0;
  size_t poped = 0;
  NodeExpanded* currentNode;
  // get straight line distance to endnode
  while (!dijkstraFinished)
  {
    // get minimum of node queue
    currentId = _openNodeQueue.top().second;
    _openNodeQueue.pop();
    poped++;

    // check if already explored
    if (_expanded[currentId] == true)
    {
       dijkstraFinished = _openNodeQueue.empty();
       continue;
    }
    currentNode = _transitGraph->getNode(currentId);
    // check if we are at final stop id
    if (currentNode->stopId == endStopId)
    {
      dijkstraFinished = true;
      continue;
    }
    // get outgoing edges from this node
    edges = _transitGraph->getOutEdges(currentId);
    // compute all new distances in loop
    for (size_t i = 0; i < edges->size(); i++)
    {
      // get all new distances
      Edge& newEdge = (*edges)[i];

      newId = newEdge.otherNodeId;

      // if node has already been explored = expanded, ignore it...
      if (_expanded[newId] == true)
        continue;
      
      newTimeFromStart = _timesFromStart[currentId] + newEdge.time;
      // compute heuristic value of new node if necessary
      if (_timesFromStart[newId] == -1 ||
          newTimeFromStart < _timesFromStart[newId])
      {
        _timesFromStart[newId] = newTimeFromStart;
        _expandedFrom[newId] = pair<int, int> (currentId, i);
        _openNodeQueue.push(std::pair<int, int> (newTimeFromStart,
                                                    newId));
      }
    }
    _expanded[currentId] = true;
    _timesFromStart[currentId] = -1;
    exploredNodes++;
    dijkstraFinished = _openNodeQueue.empty();
  }

  NodeExpanded* endNode = _transitGraph->getNode(currentId);
  double lengthOfPath = _timesFromStart[currentId];
  _timesFromStart[currentId] = -1;
  // clear queue...
  while(!_openNodeQueue.empty())
  {
    currentId = _openNodeQueue.top().second;
    _openNodeQueue.pop();
    _timesFromStart[currentId] = -1;
  }
  printf("explored: %zu\n", exploredNodes);
  printf("popped  : %zu\n", poped);
  if (endNode->stopId != endStopId)
  {
    printf("shortest Path not found from %s(%d) to %s(%d)\n",
                                _transitGraph->getStopName(startStopId).c_str(),
                                startStopId,
                                _transitGraph->getStopName(endStopId).c_str(),
                                endStopId);
    return -1;
  }
  else if (reconstructPath)
  {
    unordered_set<int> startSet;
    for (size_t i = 0; i < startNodeIds.size(); i++)
      startSet.insert(startNodeIds[i]);
    reconstructShortestPath(startSet, endNode->id);
  }
  return lengthOfPath;
}

void ExpandedDijkstra::reconstructShortestPath(const unordered_set<int>&
                                              startIds, int endId)
{
  assert(transitShortestPath.size() == 0);
  // check if dijkstra finished...
  if (_expandedFrom[endId].first == -1)
    return;
  // now insert shortest Path into vector
  // insert dummy edge for last pair
  NodeExpanded nextNode = (*(_transitGraph->getNode(endId)));
  transitShortestPath.push_back(pair<NodeExpanded, Edge>(nextNode, Edge(-1, -1)));
  int nextNodeId = endId;
  int edgeId;
  while (startIds.count(nextNodeId) == 0)
  {
    edgeId = _expandedFrom[nextNodeId].second;
    nextNodeId = _expandedFrom[nextNodeId].first;
    nextNode = (*_transitGraph->getNode(nextNodeId));
    Edge nextEdge = _transitGraph->getOutEdges(nextNodeId)->at(edgeId);
    transitShortestPath.push_back(pair<NodeExpanded, Edge>(nextNode, nextEdge));
  }
  // have to reverse the path to get proper path
  reverse(transitShortestPath.begin(), transitShortestPath.end());
}
