// Copyright 2011 Simon Skilevic und Robin Schirrmeister

#include <getopt.h>
#include <boost/asio.hpp>
#include <algorithm>
#include <ctime>
#include <iostream>
#include <string>
#include <vector>
#include <utility>
#include "./Heuristics.h"
#include "./TransitNetworkExpanded.h"
#include "./ExpandedDijkstra.h"
#include "./Timer.h"
#include "./MultiCriteriaHybridDijkstra.h"

using boost::asio::ip::tcp;
using std::vector;

// For formatting ASCII output.
#define BOLD     "\033[1m"
#define BLACK    "\033[30m"
#define RED      "\033[31m"
#define GREEN    "\033[32m"
#define BROWN    "\033[33m"
#define BLUE     "\033[34m"
#define RESET    "\033[0m"

void printUsage()
{
  printf("This program starts a server for a Hybrid Network\n"
         "(Transit plus Road Network).\n\n"
         "USAGE: ./HybridServerMain -p <Port> <GTFSFolder> <OSM-File>\n");
}

void replaceInString(string* original, const string& toReplace,
                       const string& replaceWith)
{
  size_t pos = original->find(toReplace);
  while (pos != string::npos)
  {
    original->replace(pos, toReplace.size(), replaceWith);
    pos = original->find(toReplace, pos + 1);
  }
}

using std::string;
using std::vector;
int main(int argc, char** argv)
{
  string GTFSFolder;
  string osmFile;
  // Needed for getopt_long.
  int optChr;
  optind  = 1;
  int port = 0;
  while (true)
  {
    int optionIndex = 0;
    // Define command-line options.
    static struct option longOptions[] =
    {
      {"help", no_argument, 0, 'h'},
      {"port", required_argument, 0, 'p'},
      {0, 0, 0, 0}
    };

    optChr = getopt_long(argc, argv, "hp:", longOptions, &optionIndex);

    if (optChr == -1) break;

    switch (optChr)
    {
      case 0:
        break;
      case 'h':
        printUsage();
        exit(0);
      case 'p':
        port = atoi(optarg);
        break;
      default:
        printf("unknown option: %c\n", optChr);
        printUsage();
        exit(0);
    }
  }

  printf("Using port %d...\n", port);
  // check that GTFS folder was given
  if (argc - optind != 2)
  {
    printf("Please specify GTFS input folder and OSM File!\n\n");
    printUsage();
    return 1;
  }
  GTFSFolder = argv[optind];
  osmFile = argv[optind + 1];
  // parse transit network for one day
  RoadNetwork roadNetwork;
  roadNetwork.readFromOsmFile(osmFile);
  //roadNetwork.reduceToBiggestComponent();
  TransitNetworkDependent transitNetworkDependent;
  transitNetworkDependent.parseFromGTFSFiles(GTFSFolder);
  HybridTransitRoadNetwork hybridNetwork(&roadNetwork, &transitNetworkDependent,
    NULL);
  int neighboursNumber = 4;
  hybridNetwork.buildHybridNetwork(neighboursNumber);
  MultiCriteriaHybridDijkstra routePlanner(&hybridNetwork);
  printf("Nodes in Road Network: %zu\n", roadNetwork.getNumNodes());
  printf("Nodes in Transit Network: %zu\n", transitNetworkDependent.getNumNodes());

  // Now the server part.
  try
  {
    // Create an boost IO service.
    boost::asio::io_service io_service;

    // Create socket and bind to and listen on given port.
    tcp::acceptor acceptor(io_service, tcp::endpoint(tcp::v4(), port));

    // Wait for requests and process them.
    while (true)
    {
      // Wait for request.
      std::cout << BLUE << BOLD << "Waiting for query on port " << port
                << " ... " << RESET << std::flush;
      tcp::socket socket(io_service);
      acceptor.accept(socket);
      std::time_t now = std::time(0);
      std::string daytime = std::ctime(&now); // NOLINT
      std::cout << "done, received new request on " << daytime << std::flush;

      // Get the request string.
      std::vector<char> requestBuffer(1000);
      boost::system::error_code read_error;
      socket.read_some(boost::asio::buffer(requestBuffer), read_error);
      std::string request(requestBuffer.size(), 0);
      std::copy(requestBuffer.begin(), requestBuffer.end(), request.begin());
      for (size_t i = 0; i < request.size(); i++)
        request[i] = isspace(request[i]) ? ' ' : request[i];
      std::cout << "request string is \"" << request << std::endl;

      // Extract query = the thing between the GET and the HTTP ...

      std::stringstream jsonp;  // the answer jsonp text...
      std::string query = request.substr(5);
      size_t pos = query.find(" HTTP");
      if (pos != std::string::npos) query.erase(pos);
      std::cout << "query: " << query << "\n";
      string queryCommand = "NEWQUERY:";
      string stationCommand = "GETSTATIONNAMES:";
      string resultsCommand = "GETDIJKSTRARESULTS:";
      string stepsCommand = "GETDIJKSTRASTEPS:";
      string statisticsCommand = "GETDIJKSTRASTATISTICS:";
      if (query.find(queryCommand) != string::npos)
      {
        // new query..
        query = query.substr(queryCommand.size());
        // unescape our delimiters :)
        replaceInString(&query, string("%7C%7C"), "||");
        std::cout << "query: " << query << "\n";
        size_t posStartStation = query.find("||");
        string startStation = query.substr(0, posStartStation);
        replaceInString(&startStation, string("%20"), " ");
        std::cout << "start station: " << startStation << "\n";
        size_t posEndStation = query.find("||", posStartStation + 1);
        string endStation = query.substr(posStartStation + 2,
                                         posEndStation - (posStartStation + 2));
        replaceInString(&endStation, string("%20"), " ");
        std::cout << "end station: " << endStation << "\n";
        int hour = atoi(query.c_str() + posEndStation + 2);
        size_t minutePos = query.find("||", posEndStation + 1);
        int minute = atoi(query.c_str() + minutePos + 2);
        size_t secondPos = query.find("||", minutePos + 1);
        int second = atoi(query.c_str() + secondPos + 2);
        std::cout << "Hour:   " << hour << "\n";
        std::cout << "Minute: " << minute << "\n";
        std::cout << "Second: " << second << "\n";
        Time startTime(hour, minute, second);
          printf("before computing paths\n");
        routePlanner.computeShortestPath(startStation,
                                         startTime,
                                         endStation,
                                         true);
          printf("after computing paths\n");
        if (routePlanner.optimalPaths.size() == 0)
          continue;

        jsonp << "handleNewResult(\n"
              << "{\n"
              << " \"paths\" :  [\n";
        for (size_t i = 0; i < routePlanner.optimalPaths.size(); i++)
        {
          if (i > 0) jsonp << ",\n";
          printf("before getting path\n");
          const TransitPath& curPath = routePlanner.optimalPaths[i];
          printf("before getting json\n");
          jsonp << curPath.createJSON();
          printf("after getting json\n");
          routePlanner.createAddedAndRemovedMaps();
        }
        jsonp << "  ]\n";
        jsonp << "})\n";
      }
      else if (query.find(stationCommand) != string::npos)
      {
        jsonp << "handleStationNames(\n"
              << "{ \"stopNames\" : [";
        // get all stop names
        vector<string> stopNames;
        for (size_t i = 0; i < transitNetworkDependent.getNumStops(); i++)
          stopNames.push_back(transitNetworkDependent.getStopName(i));
        for (size_t i = 0; i < stopNames.size(); i++)
        {
          jsonp << "\"" << stopNames[i] << "\"";
          if (i < stopNames.size() - 1)
            jsonp << ",";
        }
        jsonp << "]})\n";
      }
      else if (query.find(resultsCommand) != string::npos)
      {
        // function doesnt exist anymore...
        //TODO(Robin): do something else instead? heat map? :)
        /*
        assert(routePlanner.optimalPaths.size() > 0);
        jsonp << "handleDijkstraResults(\n"
              << routePlanner.createAllDijkstraLocationsJSON();
        jsonp << ")\n";*/
      }
      else if (query.find(statisticsCommand) != string::npos)
      {
        // function doesnt exist anymore...
        //TODO(Robin): do something else instead? heat map? :)

        assert(routePlanner.optimalPaths.size() > 0);
        jsonp << "handleDijkstraStatistics(\n"
              << routePlanner.createDijkstraStatisticsJSON();
        jsonp << ")\n";
      }
      else if (query.find(stepsCommand) != string::npos)
      {
        // new query..
        query = query.substr(stepsCommand.size());
        std::cout << "Labelsquery: " << query << "\n";
        int startStep = atoi(query.c_str());
        size_t endStepPos = query.find(",") + 1;
        int endStep = atoi(query.c_str() + endStepPos);
        std::cout << "Start Step: " << startStep << "\n";
        std::cout << "End Step: " << endStep << "\n";
        jsonp << "handleDijkstraSteps(\n"
              << routePlanner.createDijkstraStepsJSON(startStep, endStep);
        jsonp << ")\n";
      }
      std::ostringstream answer;
      answer << "HTTP/1.0 200 OK\r\n"
             << "Content-Length: " << jsonp.str().size() << "\r\n"
             << "Content-Type: application/javascript" << "\r\n"
             << "Connection: close\r\n"
             << "\r\n"
             << jsonp.str();
      string logText =  answer.str().substr(0, 1000) + "...";
      printf("\nAnswertext:  %s\n", logText.c_str());
      boost::system::error_code write_error;
      boost::asio::write(socket, boost::asio::buffer(answer.str()),
          boost::asio::transfer_all(), write_error);
    }
  }
  catch (std::exception& e) // NOLINT
  {
    std::cerr << e.what() << std::endl;
  }
}


