
#include <stdio.h>
#include <gtest/gtest.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unordered_map>
#include <stdlib.h>
#include <string>
#include <utility>
#include <vector>
#include "./CsvParser.h"
#include "./TransitNetworkDependent.h"
#include "./Dijkstra.h"
#include "./ExpandedDijkstra.h"

TEST(TransitNetworkDependentTest, parseStopsFile)
{
    // create stop file
  TransitNetworkDependent transitNetwork;
  string stopsFileName = "stops.txt";
  FILE* stopsFile = fopen(stopsFileName.c_str(), "w");
  fprintf(stopsFile,
    "stop_id,stop_name,stop_desc,stop_lat,stop_lon,stop_url,"
    "location_type,parent_stop\n"
    "U70,Eisenacher Strasse,Pizza :),37.728631,-122.431282,,\n"
    "U71,Bayerischer Platz,Mein Zuhause :),37.728631,-122.431282,,\n"
    "U72,Berliner Strasse, Zu U9 :),52.74103,-122.422482,,\n"
    "U73,Blissestrasse,was will man da?,37.75223,-122.418581,,\n"
    "U91,Berliner Strasse,umsteigen,37.75713,-122.418982,,\n"
    "U92,Guentzelstrasse,hmmm,37.761829,-122.419382,,\n");
  fclose(stopsFile);
  transitNetwork.parseStopsFile(stopsFileName);
  remove(stopsFileName.c_str());
  // first check that the stops have been added as nodes and
  // check their names
  ASSERT_EQ(6, transitNetwork._nodes->size());
  ASSERT_EQ("Eisenacher Strasse", transitNetwork.getStopName(0));
  ASSERT_EQ("Bayerischer Platz", transitNetwork.getStopName(1));
  ASSERT_EQ("Berliner Strasse",transitNetwork.getStopName(2));
  ASSERT_EQ("Blissestrasse", transitNetwork.getStopName(3));
  // we force stop names to be unique..
  ASSERT_EQ("Berliner Strasse U91", transitNetwork.getStopName(4));
  ASSERT_EQ("Guentzelstrasse", transitNetwork.getStopName(5));
  // now check their coordinates
  // first two nodes have same coordinates then always different ones
  ASSERT_EQ(37.728631, transitNetwork.getNode(0)->latitude);
  ASSERT_EQ(-122.431282, transitNetwork.getNode(0)->longitude);
  ASSERT_EQ(37.728631, transitNetwork.getNode(1)->latitude);
  ASSERT_EQ(-122.431282, transitNetwork.getNode(1)->longitude);
  ASSERT_EQ(52.74103, transitNetwork.getNode(2)->latitude);
  ASSERT_EQ(-122.422482, transitNetwork.getNode(2)->longitude);
  ASSERT_EQ(37.75223, transitNetwork.getNode(3)->latitude);
  ASSERT_EQ(-122.418581, transitNetwork.getNode(3)->longitude);
  ASSERT_EQ(37.75713, transitNetwork.getNode(4)->latitude);
  ASSERT_EQ(-122.418982, transitNetwork.getNode(4)->longitude);
  ASSERT_EQ(37.761829, transitNetwork.getNode(5)->latitude);
  ASSERT_EQ(-122.419382, transitNetwork.getNode(5)->longitude);
}



TEST(TransitNetworkDependentTest, parseFromGTFSFiles)
{
  TransitNetworkDependent transitNetwork;
  string testDirectory = "GTFSTempParseTestDependent";
  int couldCreateDirectory = mkdir(testDirectory.c_str(), S_IRWXU | S_IRWXG |
                                                          S_IROTH | S_IXOTH);
  if (couldCreateDirectory != 0)
  {
    printf("couldn't create testDirectory: %s\n", testDirectory.c_str());
    // make test crash if it directory could not be created...
    ASSERT_TRUE(false);
  }
  string calendarFileName = testDirectory + "/calendar.txt";
  FILE* calendarFile = fopen(calendarFileName.c_str(), "w");
  fprintf(calendarFile, "service_id,monday,tuesday,wednesday,"
                        "thursday,friday,saturday,sunday,start_date,end_date\n"
                        "WE,0,0,0,0,0,1,1,20060701,20060731\n"
                        "WD,1,1,1,1,1,0,0,20060701,20060731\n"
                        "FW,1,1,1,1,1,1,1,20060701,20060731\n");
  fclose(calendarFile);

  string tripFileName = testDirectory + "/trips.txt";
  FILE* tripFile = fopen(tripFileName.c_str(), "w");
  fprintf(tripFile, "route_id,service_id,trip_id,trip_headsign,block_id\n"
                        "A,WE,U7WE,Downtown,1\n"
                        "A,WD,U7WD,Downtown,2\n"
                        "A,FW,U7FW,Downtown,2\n"
                        "A,FW,U9FW,Downtown,2\n");
  fclose(tripFile);

  string stopsFileName = testDirectory + "/stops.txt";
  FILE* stopsFile = fopen(stopsFileName.c_str(), "w");
  fprintf(stopsFile,
    "stop_id,stop_name,stop_desc,stop_lat,stop_lon,"
    "stop_url,location_type,parent_stop\n"
    "U70,Eisenacher Strasse,Pizza :),37.728631,-122.431282,,\n"
    "U71,Bayerischer Platz,Mein Zuhause :),37.728631,-122.433282,,\n"
    "U9_U7_1,Berliner Strasse, Zu U9 :),52.74103,-122.422482,,\n"
    "U73,Blissestrasse,was will man da?,37.75223,-122.418581,,\n"
    "U92,Guentzelstrasse,hmmm,37.761829,-122.419382,,\n");
  fclose(stopsFile);

  string frequenciesFileName = testDirectory + "/frequencies.txt";
  FILE* frequenciesFile = fopen(frequenciesFileName.c_str(), "w");
  fprintf(frequenciesFile,
    "trip_id,start_time,end_time,headway_secs\n"
    "U9FW,05:30:03,26:30:03,10800\n"
    "U7WE,00:00:00,24:00:00,5400\n"
    "U7WD,05:29:30,26:30:00,3600\n"
    "U7FW,05:30:00,25:30:00,7200\n");
  fclose(frequenciesFile);

  string stopTimesFileName = testDirectory + "/stop_times.txt";
  FILE* stopTimesFile = fopen(stopTimesFileName.c_str(), "w");
  fprintf(stopTimesFile,
    "trip_id,arrival_time,departure_time,stop_id,"
    "stop_sequence,pickup_type,dropoff_type\n"
    "U7FW,5:30:0,5:30:0,U70,1,0,0,0\n"
    "U7FW,5:32:2,5:33:0,U71,2,0,0,0\n"
    "U7FW,5:34:0,5:35:0,U9_U7_1,3,0,0,0\n"
    "U7FW,5:36:0,5:37:0,U73,4,0,0,0\n"
    "U7WE,0:0:0,0:1:0,U71,1,0,0,0\n"
    "U7WE,0:4:0,0:6:0,U9_U7_1,2,0,0,0\n"
    "U7WD,5:29:30,05:31:02,U70,1,0,0,0\n"
    "U7WD,5:32:0,05:33:20,U71,2,0,0,0\n"
    "U7WD,5:34:0,05:35:0,U9_U7_1,3,0,0,0\n"
    "U9FW,5:30:03,5:32:0,U9_U7_1,1,0,0,0\n"
    "U9FW,5:34:0,05:38:0,U92,2,0,0,0\n");
  // last stop transfer and departure time equal if waiting time = 4 minutes
  fclose(stopTimesFile);
  transitNetwork.parseFromGTFSFiles(testDirectory);
  // remove files and directory...
  remove(calendarFileName.c_str());
  remove(tripFileName.c_str());
  remove(stopsFileName.c_str());
  remove(frequenciesFileName.c_str());
  remove(stopTimesFileName.c_str());
  int couldRemoveDirectory = rmdir(testDirectory.c_str());
  if (couldRemoveDirectory != 0)
    printf("couldn't remove directory %s\n", testDirectory.c_str());

  // Check that all stops are correctly added as nodes
  ASSERT_EQ(5, transitNetwork.getNumNodes());
  ASSERT_EQ(0, transitNetwork.getNodeIdByTheStopName("Eisenacher Strasse"));
  ASSERT_EQ(1, transitNetwork.getNodeIdByTheStopName("Bayerischer Platz"));
  ASSERT_EQ(2, transitNetwork.getNodeIdByTheStopName("Berliner Strasse"));
  ASSERT_EQ(3, transitNetwork.getNodeIdByTheStopName("Blissestrasse"));
  ASSERT_EQ(4, transitNetwork.getNodeIdByTheStopName("Guentzelstrasse"));

  // Check that coordinates are correct
  ASSERT_EQ(37.728631, transitNetwork.getNode(0)->latitude);
  ASSERT_EQ(-122.431282, transitNetwork.getNode(0)->longitude);
  ASSERT_EQ(37.728631, transitNetwork.getNode(1)->latitude);
  ASSERT_EQ(-122.433282, transitNetwork.getNode(1)->longitude);
  ASSERT_EQ(52.74103, transitNetwork.getNode(2)->latitude);
  ASSERT_EQ(-122.422482, transitNetwork.getNode(2)->longitude);
  ASSERT_EQ(37.75223, transitNetwork.getNode(3)->latitude);
  ASSERT_EQ(-122.418581, transitNetwork.getNode(3)->longitude);
  ASSERT_EQ(37.761829, transitNetwork.getNode(4)->latitude);
  ASSERT_EQ(-122.419382, transitNetwork.getNode(4)->longitude);
  // check that each trip available on monday exists by checking characteristic
  // edges

  // Check that Eisenacher Strasse has correct edges to Bayerischer Platz
  vector<EdgeDependent>* outEdges = transitNetwork.getOutEdges(0);
  // check first few connections explicitly, then check all remaining times
  // approximately (only check possibilities for minutes and seconds)
  // then check endtimes explicitly
  ASSERT_EQ(1, outEdges->size());
  ASSERT_EQ(1, outEdges->at(0).otherNodeId);
  const vector<Connection>& times = outEdges->at(0)._times;
  ASSERT_EQ(Time(0, 31, 2), times[0].departureTime);
  ASSERT_EQ(Time(0, 32, 0), times[0].arrivalTime);
  ASSERT_EQ(Time(1, 30, 0), times[1].departureTime);
  ASSERT_EQ(Time(1, 32, 2), times[1].arrivalTime);
  ASSERT_EQ(Time(1, 31, 2), times[2].departureTime);
  ASSERT_EQ(Time(1, 32, 0), times[2].arrivalTime);
  ASSERT_EQ(Time(2, 31, 2), times[3].departureTime);
  ASSERT_EQ(Time(2, 32, 0), times[3].arrivalTime);
  ASSERT_EQ(Time(5, 30, 0), times[4].departureTime);
  ASSERT_EQ(Time(5, 32, 2), times[4].arrivalTime);
  for (size_t i = 5; i < times.size(); i++)
  {
    ASSERT_TRUE(times[i].departureTime.minute() == 30 ||
                times[i].departureTime.minute() == 31);
    ASSERT_TRUE(times[i].departureTime.second() == 0 ||
                times[i].departureTime.second() == 2);
    ASSERT_EQ(32, times[i].arrivalTime.minute());
    ASSERT_TRUE(times[i].arrivalTime.second() == 0 ||
                times[i].arrivalTime.second() == 2);
  }
  ASSERT_TRUE(times[times.size() - 2].departureTime == Time(23, 30, 0));
  ASSERT_TRUE(times[times.size() - 2].arrivalTime == Time(23, 32, 2));
  ASSERT_TRUE(times[times.size() - 1].departureTime == Time(23, 31, 2));
  ASSERT_TRUE(times[times.size() - 1].arrivalTime == Time(23, 32, 0));

  // check that there are two edges from berliner strasse, to blissestrasse
  // and to guentzelstrasse
  outEdges = transitNetwork.getOutEdges(2);
  // check first few connections explicitly, then check all remaining times
  // approximately (only check possibilities for minutes and seconds)
  // then check endtimes explicitly
  ASSERT_EQ(2, outEdges->size());
  ASSERT_EQ(4, outEdges->at(0).otherNodeId);
  ASSERT_EQ(3, outEdges->at(1).otherNodeId);


  // Check U9 trip to Guentzelstrasse
  const vector<Connection>& times2 = outEdges->at(0)._times;
  ASSERT_EQ(Time(2, 32, 0), times2[0].departureTime);
  ASSERT_EQ(Time(2, 34, 0), times2[0].arrivalTime);
  size_t i = 1;
  for (int hour = 5; hour < 24; hour += 3)
  {
    ASSERT_EQ(Time(hour, 32, 0), times2[i].departureTime);
    ASSERT_EQ(Time(hour, 34, 0), times2[i].arrivalTime);
    i++;
  }

  // Check U7 trip to Blissestrasse
  const vector<Connection>& times3 = outEdges->at(1)._times;
  ASSERT_EQ(Time(1, 35, 0), times3[0].departureTime);
  ASSERT_EQ(Time(1, 36, 0), times3[0].arrivalTime);
  i = 1;
  for (int hour = 5; hour < 24; hour += 2)
  {
    ASSERT_EQ(Time(hour, 35, 0), times3[i].departureTime);
    ASSERT_EQ(Time(hour, 36, 0), times3[i].arrivalTime);
    i++;
  }
}

TEST(HybridNetworkTest, getConnections)
{
  // Testing transit network dependent get connections...
  {
    TransitNetworkDependent transitNetwork;

    string testDirectory = "./GTFSMultiDependentTest";
    int status = mkdir(testDirectory.c_str(), S_IRWXU | S_IRWXG | S_IROTH |
                                              S_IXOTH);
    if (status != 0)
    {
      printf("couldn't create testDirectory: %s\n", testDirectory.c_str());
      // make test crash if it directory could not be created...
      ASSERT_TRUE(false);
    }
    string calendarFileName =  testDirectory + "/calendar.txt";
    FILE* calendarFile = fopen(calendarFileName.c_str(), "w");
    fprintf(calendarFile, "service_id,monday,tuesday,wednesday,thursday,"
                          "friday,saturday,sunday,start_date,end_date\n"
                          "WE,0,0,0,0,0,1,1,20060701,20060731\n"
                          "WD,1,1,1,1,1,0,0,20060701,20060731\n"
                          "FW,1,1,1,1,1,1,1,20060701,20060731\n");
    fclose(calendarFile);

    string tripFileName =  testDirectory + "/trips.txt";
    FILE* tripFile = fopen(tripFileName.c_str(), "w");
    fprintf(tripFile, "route_id,service_id,trip_id,trip_headsign,block_id\n"
                          "A,WE,U7WE,Downtown,1\n"
                          "A,WD,U7WD,Downtown,2\n"
                          "A,FW,U7FW,Downtown,2\n"
                          "A,FW,U9FW,Downtown,2\n"
                          "A,FW,SHORTCUTTRANSFER,Downtown,2\n"
                          "A,WD,SHORTCUTTIME,Downtown,2\n");  // add extra trips
                                                       // for transfer tests...
    fclose(tripFile);

    string stopsFileName = testDirectory + "/stops.txt";
    FILE* stopsFile = fopen(stopsFileName.c_str(), "w");
    fprintf(stopsFile,
      "stop_id,stop_name,stop_desc,stop_lat,stop_lon,stop_url,"
      "location_type,parent_stop\n"
      "U70,Eisenacher Strasse,Pizza :),37.728631,-122.431282,,\n"
      "U71,Bayerischer Platz,Mein Zuhause :),37.728631,-122.433282,,\n"
      "U9_U7_1,Berliner Strasse, Zu U9 :),52.74103,-122.422482,,\n"
      "U73,Blissestrasse,was will man da?,37.75223,-122.418581,,\n"
      "U92,Guentzelstrasse,hmmm,37.761829,-122.419382,,\n");
    fclose(stopsFile);

    string frequenciesFileName = testDirectory + "/frequencies.txt";
    FILE* frequenciesFile = fopen(frequenciesFileName.c_str(), "w");
    fprintf(frequenciesFile,
      "trip_id,start_time,end_time,headway_secs\n"
      "U9FW,05:30:00,25:30:00,10800\n"
      "U7WE,00:00:00,24:00:00,5400\n"
      "U7WD,05:26:00,26:30:00,3600\n"
      "U7FW,05:29:32,25:00:00,7200\n"
      "SHORTCUTTRANSFER,05:29:32,25:00:00,7200\n"
      "SHORTCUTTIME,05:29:32,25:00:00,7200\n");
    fclose(frequenciesFile);

    string stopTimesFileName = testDirectory + "/stop_times.txt";
    FILE* stopTimesFile = fopen(stopTimesFileName.c_str(), "w");
    fprintf(stopTimesFile,
      "trip_id,arrival_time,departure_time,stop_id,stop_sequence,"
      "pickup_type,dropoff_type\n"
      "U9FW,5:30:0,5:32:17,U9_U7_1,1,0,0,0\n"
      "U9FW,5:34:13,05:38:0,U92,2,0,0,0\n"
      // last stop transfer and departure time equal if transfer time = 4 minutes
      "U7WE,0:0:0,0:1:0,U71,1,0,0,0\n"
      "U7WE,0:4:0,0:6:0,U9_U7_1,2,0,0,0\n"
      "U7WD,5:26:0,05:26:32,U70,1,0,0,0\n"
      "U7WD,5:26:58,05:27:10,U71,2,0,0,0\n"
      "U7WD,5:27:35,05:29:0,U9_U7_1,3,0,0,0\n"
      "U7FW,5:29:32,5:30:0,U70,1,0,0,0\n"
      "U7FW,5:32:0,5:33:0,U71,2,0,0,0\n"
      "U7FW,5:34:0,5:35:0,U9_U7_1,3,0,0,0\n"
      "U7FW,5:36:0,5:37:0,U73,4,0,0,0\n"
      "SHORTCUTTRANSFER,5:29:32,5:30:0,U71,1,0,0,0\n"
      "SHORTCUTTRANSFER,5:38:0,5:40:0,U92,2,0,0,0\n"
      "SHORTCUTTIME,5:29:32,5:30:58,U71,1,0,0,0\n"
      "SHORTCUTTIME,5:32:02,5:32:0,U73,2,0,0,0\n");
    fclose(stopTimesFile);

    transitNetwork.parseFromGTFSFiles(testDirectory);

    remove(calendarFileName.c_str());
    remove(tripFileName.c_str());
    remove(stopsFileName.c_str());
    remove(frequenciesFileName.c_str());
    remove(stopTimesFileName.c_str());
    status = rmdir(testDirectory.c_str());
    if (status != 0)
      printf("couldn't remove directory %s\n", testDirectory.c_str());
    bool isMultiCriteria = true;
    bool isStartNode = true;
    int nodeId = transitNetwork.getNodeIdByTheStopName("Bayerischer Platz");
    // Early test: Test at start of service whether all line ids are found
    Time arrivalTime(5, 26, 0);
    int tripId = -1;  // Test getting all connections for all trips by using
                      // invalid trip id...
    vector<LiveConnection> outConnections;
    transitNetwork.getConnections(nodeId, arrivalTime, tripId,
                                  isMultiCriteria, isStartNode,
                                  &outConnections);
    ASSERT_EQ(4, outConnections.size());
    // WARNING: Test assumes a certain order of trips that could
    // actually change (without affecting correctness)
    ASSERT_EQ(Time(5, 27, 10), outConnections[0].connection.departureTime);
    ASSERT_EQ(Time(5, 27, 35), outConnections[0].connection.arrivalTime);
    ASSERT_EQ(transitNetwork.getNodeIdByTheStopName("Berliner Strasse"),
              outConnections[0].toNodeId);
    ASSERT_EQ(0, outConnections[0].transfers);
  }
}
 