// Copyright November 18, 2011 Robin Schirrmeister and Simon Skilevic
// Chair of Algorithms and Datastructures Uni Freiburg


#include "./TransitParser.h"
#include <utility>
#include <string>
#include <vector>

using std::pair;

TransitParser::TripNode::TripNode(const string& GTFSTimeString,
                                         int stopId,
                                         NodeType type)
{
  char hour = atoi(GTFSTimeString.c_str());
  int posFirstColon = GTFSTimeString.find(':');
  char minutes = atoi(GTFSTimeString.c_str() + posFirstColon + 1);
  int posSecondColon = GTFSTimeString.find(':', posFirstColon + 1);
  char seconds = atoi(GTFSTimeString.c_str() + posSecondColon + 1);
  this->secondsSinceDay = 60 * 60 * hour + 60 * minutes + seconds;
  this->stopId = stopId;
  this->type = type;
}
TransitParser::TripFrequency::TripFrequency()
{
      intervalInSeconds = -1;
      startHour = -1;
      startMinute = -1;
      startSecond = -1;
      endHour = -1;
      endMinute = -1;
      endSecond = -1;
}

TransitParser::TripFrequency::TripFrequency(char startHour,
                                                   char startMinute,
                                                   char startSecond,
                                                   char endHour,
                                                   char endMinute,
                                                   char endSecond,
                                                   int intervalInSeconds)
{
      this->startHour = startHour;
      this->startMinute = startMinute;
      this->startSecond = startSecond;
      this->endHour = endHour;
      this->endMinute = endMinute;
      this->endSecond = endSecond;
      this->intervalInSeconds = intervalInSeconds;
}

Stop::Stop(int stopId, const string& stopName,
                           double latitude, double longitude)
{
      this->stopId = stopId;
      this->stopName = stopName;
      this->latitude = latitude;
      this->longitude = longitude;
}


TransitParser::TransitParser(int verbosity)
{
  _verbosity = verbosity;
}


// _____________________________________________________________________________
void TransitParser::parseCalendarFile(const string& calendarFileName,
                                       char validWeekDay,
                                       unordered_set<string>* validServiceIds)
{
  assert(-1 < validWeekDay && validWeekDay < 7);
  // test if file exists
  FILE* calendarFile = fopen(calendarFileName.c_str(), "r");
  if (calendarFile == NULL)
  {
    if (_verbosity > 0)
      printf("  %s does not exist...\n", calendarFileName.c_str());
    return;
  }
  fclose(calendarFile);
  CsvParser calendarParser;
  calendarParser.openFile(calendarFileName);
  int serviceIdColumn  = -1;
  int mondayColumn= -1;
  int tuesdayColumn= -1;
  int wednesdayColumn = -1;
  int thursdayColumn = -1;
  int fridayColumn = -1;
  int saturdayColumn = -1;
  int sundayColumn = -1;
  for (size_t i = 0; i < calendarParser.getNumColumns(); i++)
  {
    if (string(calendarParser.getItem(i)) == "service_id")
      serviceIdColumn = i;
    else if (string(calendarParser.getItem(i)) == "monday")
      mondayColumn = i;
    else if (string(calendarParser.getItem(i)) == "tuesday")
      tuesdayColumn = i;
    else if (string(calendarParser.getItem(i)) == "wednesday")
      wednesdayColumn = i;
    else if (string(calendarParser.getItem(i)) == "thursday")
      thursdayColumn = i;
    else if (string(calendarParser.getItem(i)) == "friday")
      fridayColumn = i;
    else if (string(calendarParser.getItem(i)) == "saturday")
      saturdayColumn = i;
    else if (string(calendarParser.getItem(i)) == "sunday")
      sundayColumn = i;
  }
  calendarParser.readNextNonEmptyLine();
  // Remember which service ids are valid on the given weekday....
  bool weekDayIsValid;
  while (!calendarParser.eof())
  {
    string serviceId = calendarParser.getItem(serviceIdColumn);

    switch (validWeekDay)
    {
      case 0:
        weekDayIsValid = calendarParser.getItem(mondayColumn) == "1";
        break;
      case 1:
        weekDayIsValid = calendarParser.getItem(tuesdayColumn) == "1";
        break;
      case 2:
        weekDayIsValid = calendarParser.getItem(wednesdayColumn) == "1";
        break;
      case 3:
        weekDayIsValid = calendarParser.getItem(thursdayColumn) == "1";
        break;
      case 4:
        weekDayIsValid = calendarParser.getItem(fridayColumn) == "1";
        break;
      case 5:
        weekDayIsValid = calendarParser.getItem(saturdayColumn) == "1";
        break;
      case 6:
        weekDayIsValid = calendarParser.getItem(sundayColumn) == "1";
        break;
      default:
        if (_verbosity > 0)
          printf("unknown weekday: %d\n", validWeekDay);
        return;
    }
    if (weekDayIsValid)
      validServiceIds->insert(serviceId);
    calendarParser.readNextNonEmptyLine();
  }
}
// _____________________________________________________________________________
void TransitParser::parseTripsFile(const string& tripFileName,
                                   const unordered_set<string>&
                                                          validServiceIds,
                                 unordered_map<string, int>* tripIdNameToTripId)
{
  // test if file exists
  FILE* tripFile = fopen(tripFileName.c_str(), "r");
  if (tripFile == NULL)
  {
    if (_verbosity > 0)
      printf("  %s does not exist...\n", tripFileName.c_str());
    return;
  }
  fclose(tripFile);
  CsvParser tripParser;
  tripParser.openFile(tripFileName);
  int serviceIdColumn  = -1;
  int tripIdColumn= -1;
  for (size_t i = 0; i < tripParser.getNumColumns(); i++)
  {
    if (string(tripParser.getItem(i)) == "service_id")
      serviceIdColumn = i;
    else if (string(tripParser.getItem(i)) == "trip_id")
      tripIdColumn = i;
  }

  tripParser.readNextLine();
  size_t tripIdInteger = 0;
  while (!tripParser.eof())
  {
    string serviceId = tripParser.getItem(serviceIdColumn);
    string tripId = tripParser.getItem(tripIdColumn);
    // check if trip id has a valid service id, if yes add it to valid trip ids
    // and give it the next trip id...
    if (validServiceIds.count(serviceId) != 0)
    {
      tripIdNameToTripId->insert(pair<string, int>(tripId, tripIdInteger));
    }
    // always increment so that trip ids correspond to lines in the file
    tripIdInteger++;
    tripParser.readNextLine();
  }
}

// _____________________________________________________________________________
void TransitParser::parseStopsFile(const string& stopFileName,
                                    vector<Stop>* stops,
                                    unordered_map<string, int>* stopIdNameToId)
{  // test if file exists
  FILE* stopFile = fopen(stopFileName.c_str(), "r");
  if (stopFile == NULL)
  {
    if (_verbosity > 0)
      printf("  %s does not exist...\n", stopFileName.c_str());
    return;
  }
  fclose(stopFile);
  CsvParser stopParser;
  stopParser.openFile(stopFileName);
  int stopIdNameColumn  = -1;
  int stopNameColumn = -1;
  int stopLatitudeColumn = -1;
  int stopLongitudeColumn = -1;
  for (size_t i = 0; i < stopParser.getNumColumns(); i++)
  {
    if (string(stopParser.getItem(i)) == "stop_id")
      stopIdNameColumn = i;
    else if (string(stopParser.getItem(i)) == "stop_name")
      stopNameColumn = i;
    else if (string(stopParser.getItem(i)) == "stop_lat")
      stopLatitudeColumn = i;
    else if (string(stopParser.getItem(i)) == "stop_lon")
      stopLongitudeColumn = i;
  }
  stopParser.readNextLine();
  int stopId = 0;
  // force stop names to be unique for safety...
  unordered_set<string> stopNames;
  while (!stopParser.eof())
  {
    string stopIdName = stopParser.getItem(stopIdNameColumn);
    // stop ids shoud be unique!
    if (stopIdNameToId->count(stopIdName) != 0)
    {
      if (_verbosity > 0)
        printf("ignoring duplicate stop id %s\n", stopIdName.c_str());
      stopParser.readNextLine();
      continue;
    }
    (*stopIdNameToId)[stopIdName] = stopId;
    string stopName = stopParser.getItem(stopNameColumn);

    // just add stopid name to make stop name unique...
    while (stopNames.count(stopName) != 0)
      stopName += string(" ") + stopIdName;
    stopNames.insert(stopName);
    double latitude = atof(stopParser.getItem(stopLatitudeColumn).c_str());
    double longitude = atof(stopParser.getItem(stopLongitudeColumn).c_str());
    Stop stop(stopId, stopName, latitude, longitude);
    stops->push_back(stop);
    stopParser.readNextLine();
    stopId++;
  }
}


// _____________________________________________________________________________
void TransitParser::parseFrequenciesFile(const string& frequencyFileName,
                          unordered_map<string, vector<TripFrequency> >*
                          tripIdNameToFrequencies)
{
  // Test if file exists
  FILE* frequenciesFile = fopen(frequencyFileName.c_str(), "r");
  if (frequenciesFile == NULL)
  {
    if (_verbosity > 0)
      printf("  file %s doesn't exist\n", frequencyFileName.c_str());
    return;
  }
  fclose(frequenciesFile);
  CsvParser frequenciesParser;
  frequenciesParser.openFile(frequencyFileName);
  int tripIdColumn  = -1;
  int startTimeColumn = -1;
  int endTimeColumn = -1;
  int headWayColumn = -1;

  for (size_t i = 0; i < frequenciesParser.getNumColumns(); i++)
  {
    if (string(frequenciesParser.getItem(i)) == "trip_id")
      tripIdColumn = i;
    else if (string(frequenciesParser.getItem(i)) == "start_time")
      startTimeColumn = i;
    else if (string(frequenciesParser.getItem(i)) == "end_time")
      endTimeColumn = i;
    else if (string(frequenciesParser.getItem(i)) == "headway_secs")
      headWayColumn = i;
  }
  frequenciesParser.readNextNonEmptyLine();
  while (!frequenciesParser.eof())
  {
    string tripId = frequenciesParser.getItem(tripIdColumn);
    string startString = frequenciesParser.getItem(startTimeColumn);
    string endString = frequenciesParser.getItem(endTimeColumn);
    int frequencyInSeconds = atoi(frequenciesParser.getItem(headWayColumn)
                                  .c_str());
    // parse start and end times of this frequency...
    char startHour = atoi(startString.c_str());
    // minutes follow after first ":"
    int posFirstColon = startString.find(':');
    char startMinute = atoi(startString.c_str() + posFirstColon + 1);
    // seconds follow after second ":"
    int posSecondColon = startString.find(':', posFirstColon + 1);
    char startSecond = atoi(startString.c_str() + posSecondColon + 1);
    char endHour = atoi(endString.c_str());
    // minutes follow after first ":"
    posFirstColon = endString.find(':');
    char endMinute = atoi(endString.c_str() + posFirstColon + 1);
    // seconds follow after second ":"
    posSecondColon = endString.find(':', posFirstColon + 1);
    char endSecond = atoi(endString.c_str() + posSecondColon + 1);
    // add new frequencies to frequencies for this trip id
    TripFrequency frequency(startHour, startMinute, startSecond,
                            endHour, endMinute, endSecond,
                            frequencyInSeconds);
    (*tripIdNameToFrequencies)[tripId].push_back(frequency);
    frequenciesParser.readNextNonEmptyLine();
  }
}

// _____________________________________________________________________________
void TransitParser::parseStopTimesFile(const string& stopTimesFileName,
        const unordered_map<string, int>& stopIdNameToId,
        const unordered_map<string, vector<TripFrequency> >&
                                          tripIdNameToFrequencies,
        const unordered_map<string, int>& tripIdNameToTripId,
        int transferTimeInSeconds,
        vector<Trip>* parsedTrips)
{
  // test if file exists
  FILE* stopTimesFile = fopen(stopTimesFileName.c_str(), "r");
  if (stopTimesFile == NULL)
  {
    if (_verbosity > 0)
      printf("  file %s doesn't exist\n", stopTimesFileName.c_str());
    return;
  }
  fclose(stopTimesFile);
  // TODO(Robin): check if sorted file already exists, if yes (and newer
  // than stop times file), just use it
  if (_verbosity > 1)
    printf("  sorting file...");
  fflush(stdout);
  // first sort file so that trips are in one block, sorted by stop sequence..
  CsvParser stopTimesParser;
  stopTimesParser.openFile(stopTimesFileName.c_str());
  int tripIdColumn  = -1;
  int arrivalTimeColumn = -1;
  int departureTimeColumn = -1;
  int stopIdColumn = -1;
  int stopSequenceColumn = -1;
  for (size_t i = 0; i < stopTimesParser.getNumColumns(); i++)
  {
    if (string(stopTimesParser.getItem(i)) == "trip_id")
      tripIdColumn = i;
    else if (string(stopTimesParser.getItem(i)) == "arrival_time")
      arrivalTimeColumn = i;
    else if (string(stopTimesParser.getItem(i)) == "departure_time")
      departureTimeColumn = i;
    else if (string(stopTimesParser.getItem(i)) == "stop_id")
      stopIdColumn = i;
    else if (string(stopTimesParser.getItem(i)) == "stop_sequence")
      stopSequenceColumn = i;
  }
  stopTimesParser.closeFile();
  string sortedFileName = stopTimesFileName + string(".sorted");
  // ignore first line (headers), then sort rest of file
  std::ostringstream sortCommandStream;
  sortCommandStream << "more +2 " << stopTimesFileName << " | "
                    << "sort -t, -k" << tripIdColumn + 1 << ","
                    << tripIdColumn + 1 << "r "
                    << "-k" << stopSequenceColumn + 1 << ","
                    << stopSequenceColumn + 1 << "n >> "
                    << sortedFileName;
  string sortCommand = sortCommandStream.str();
  int sortResult = system(sortCommand.c_str());
  if (_verbosity > 1)
    printf("  done.\n");
  if (sortResult != 0)
  {
    if (_verbosity > 0)
      printf("Couldn't sort %s into %s with command:\n %s\n",
             stopTimesFileName.c_str(),
             sortedFileName.c_str(),
             sortCommand.c_str());
    stopTimesParser.openFile(stopTimesFileName);
    stopTimesParser.readNextLine();
  }
  else
  {
    stopTimesParser.openFile(sortedFileName);
  }
  if (_verbosity > 1)
    printf("  adding nodes...");
  fflush(stdout);
  // Now continue with sorted file:
  // Get all nodes for a trip,
  // Add trip with frequency and weekdays
  // So parse the file trip by trip, also continually update which stop
  // has which nodes of the time-expanded network
  vector<vector<int> > stopIdsToNodeIds;
  stopIdsToNodeIds.resize(stopIdNameToId.size());
  // now look for the tripIdString, as long as its the same you are working
  // on the same trip..
  string tripIdName = string(stopTimesParser.getItem(tripIdColumn));
  string newTripIdName;
  bool doneParsing = false;
  bool atStartOfLoop = true;
  Trip currentTrip;
  while (!doneParsing)
  {
    bool newValidTripFound = false;
    // read one line
    if (!atStartOfLoop)
      stopTimesParser.readNextLine();
    else
      atStartOfLoop = false;
    // check if end of file has been reached
    if (stopTimesParser.eof())
      doneParsing = true;
     // ignore empty lines
    if (!doneParsing && stopTimesParser.getNumColumns() == 0)
    {
      continue;
    }


    // check if new trip is a valid trip, if not ignore this line...
    if (!doneParsing)
    {
      newTripIdName = string(stopTimesParser.getItem(tripIdColumn));
      if (tripIdNameToTripId.count(newTripIdName) == 0)
      {
        // ignore invalid trips (not running on the valid day of the network)
        continue;
      }
      if (newTripIdName != tripIdName)
        newValidTripFound = true;
    }
    // if we have a new valid trip, remember its name
    if (currentTrip.nodes.size() == 0 && newValidTripFound)
      tripIdName = newTripIdName;
    // if we have a trip with size bigger 0 and
    // we either reached the end of the file or
    // we found a new valid trip id, add the old trip...
    else if (currentTrip.nodes.size() > 0 && (doneParsing || newValidTripFound))
    {
      // get all frequencies for this trip...
      vector<TripFrequency>& frequencies = currentTrip.frequencies;
      if (tripIdNameToFrequencies.count(tripIdName) != 0)
      {
        frequencies = tripIdNameToFrequencies.find(tripIdName)->second;
      }
      else
      {
        // insert dummy frequency
        char startHour = currentTrip.nodes[0].secondsSinceDay / (60 * 60);
        char startMinute = (currentTrip.nodes[0].secondsSinceDay % 3600) / 60;
        char startSecond = currentTrip.nodes[0].secondsSinceDay % 60;
        char endHour = currentTrip.nodes.back().secondsSinceDay / (60 * 60);
        char endMinute = (currentTrip.nodes.back().secondsSinceDay % 3600) / 60;
        char endSecond = currentTrip.nodes.back().secondsSinceDay % 60;
        TripFrequency dummyFrequency(startHour, startMinute, startSecond,
                                     endHour, endMinute, endSecond,
                                     -1);
        frequencies.push_back(dummyFrequency);
      }
      int tripId = tripIdNameToTripId.find(tripIdName)->second;
      currentTrip.tripId = tripId;
      // add current trip and go for next trip...
      parsedTrips->push_back(currentTrip);
      currentTrip = Trip();
      if (!doneParsing)
        tripIdName = newTripIdName;
    }
    // now add the new stop times from this line to the current trip...
    if (!doneParsing)
    {
      string arrivalTimeString(stopTimesParser.getItem(arrivalTimeColumn));
      if (arrivalTimeString == "")
      {
        // ignore this line
        if (_verbosity > 0)
              printf("Arrival Time not given for trip idname %s\n",
                     tripIdName.c_str());
        continue;
      }
      string departureTimeString(stopTimesParser.getItem(departureTimeColumn));
      if (departureTimeString == "")
      {
        if (_verbosity > 0)
          printf("Departure Time not given for trip idname %s\n",
                                                   tripIdName.c_str());
        continue;
      }
      string stopIdName(stopTimesParser.getItem(stopIdColumn));
      if (stopIdNameToId.count(stopIdName) == 0)
      {
        if (_verbosity > 0)
          printf("Warning: stop id %s not found\n", stopIdName.c_str());
        continue;
      }
      int stopId = stopIdNameToId.find(stopIdName)->second;
      // always add nodes in order: arrival-transfer-departure to the trip
      TripNode arrivalNode(arrivalTimeString, stopId, ARRIVALNODE);
      TripNode transferNode(arrivalTimeString, stopId, TRANSFERNODE);
      transferNode.secondsSinceDay += transferTimeInSeconds;
      TripNode departureNode(departureTimeString, stopId, DEPARTURENODE);
      currentTrip.nodes.push_back(arrivalNode);
      currentTrip.nodes.push_back(transferNode);
      currentTrip.nodes.push_back(departureNode);
    }
  }
  // now remove sorted file
  remove(sortedFileName.c_str());
}
